///////////////////////////////////////////////////////////////////////////////
//
// IBM VisualAge C++
// Licensed Materials - Property of IBM
//
// Copyright IBM Corp. 1998,2000. All Rights Reserved
//
// DISCLAIMER:
// This sample code is subject to the license and terms and conditions 
// found in the license.txt file in this directory. It is not part of any 
// standard IBM product and is provided to you solely for the purpose of 
// assisting you in the development of your applications. The code is 
// provided 'AS IS', without warranty or condition of any kind. IBM shall 
// not be liable for any damages arising out of your use of the sample code, 
// even if advised of the possibility of such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "metrics.h"
#include <services/Array.t>


//CS_TraceBackInformation::CS_TraceBackInformation(struct sigcontext * a){}

//////////////////////////////////////////////////////////////////////////////
//
// Function Name : numberOfLinesOfCode
// Description   : Computes the number of lines of code for a give source file.
//               : It iterates through the token streams for each source region,
//               : counting the number of preprocessed lines of code.
//               : 
// Input         : ps - the primary source file
//               :        
// Output        : None
// Returns       : number lines of pre-processed code in the source file
//
unsigned int MetricsCalculator::numberOfLinesOfCode(CS_PrimarySource& ps)
   {
   CS_StorageArena arena;
   CS_StoragePlacement placement(&arena);
   CS_SourceRegionIterator &i = ps.sourceRegionIterator(placement);
   unsigned int linesOfCode = 0;
   unsigned int currentLine;
   unsigned int nextLine;
   CS_Token token;
 
   for (CS_SourceRegion *sourceRegion = i.first(); sourceRegion; sourceRegion = i.next())
      {
      CS_TokenLocation firstTokenLocation = sourceRegion->tokenLocation();
      CS_TokenIterator tokenIterator(firstTokenLocation, CS_true);
      currentLine = firstTokenLocation.line();

      // Iterate through the tokens in the source region
      for (;;)
         {
         tokenIterator.get(&token);
         if (token.kind() == CS_Token::TK_EOF)
            {
            break;
            }

         CS_TokenLocation &location = token.tokenLocation();
         nextLine = location.line();
   
         if (currentLine != nextLine)
            {
            linesOfCode++;
            currentLine = nextLine;
            }
         }
      }
   return linesOfCode;
   } 

//////////////////////////////////////////////////////////////////////////////
//
// Function Name : numberOfMemberFunctions
// Description   : Computes the number of member functions in a class.
//               : It iterates through the member declaration store checking that
//               : a declaration is a function and maps to a real source location
//               : 
// Input         : classDecl - any class declaration
//               :        
// Output        : None
// Returns       : number of member functions in the class
//

unsigned int MetricsCalculator::numberOfMemberFunctions(CS_ClassDeclaration& classDecl)
   {
   CS_MemberDeclarationStore& memberDeclStore = classDecl.memberDeclarationStore();
   unsigned int numberOfMemberFunctions = 0;

   // TO DO: handle nested classes
   for (CS_Declaration* memberDecl = memberDeclStore.firstDeclaration();
        memberDecl;
        memberDecl = memberDeclStore.nextDeclaration(*memberDecl))
      {
      if (memberDecl->asFunctionDeclaration() && memberDecl->mapsToASourceLocation())
         numberOfMemberFunctions++;
      }
   return numberOfMemberFunctions;
   }       

//////////////////////////////////////////////////////////////////////////////
//
// Function Name : depthOfInheritanceTree
// Description   : Computes the depth of an inheritance tree for a class.
//               : It uses a class' base specifier to trace ancestory.
//               : Currently does not traverse all parents, only the first.
//               : 
// Input         : decl - any class declaration
//               :        
// Output        : None
// Returns       : depth of inheritance tree
//
unsigned int MetricsCalculator::depthOfInheritanceTree(CS_ClassDeclaration& decl)
   {
   CS_BaseSpecifier* base = decl.firstBaseSpecifier();
   
   // TO DO: Handle multiple inheritance. Probably some funky recursion needed!
   unsigned int depth = 0;
   for (;;)
      {
      if (!base)
        break;

      depth++;
      CS_ClassDeclaration& classDecl = base->classDeclaration();
      base = classDecl.firstBaseSpecifier();

      }
   return depth;
   }

//////////////////////////////////////////////////////////////////////////////
//
// Function Name : numberOfChildren
// Description   : Computes the number of children for a class.
//               : This information is contained in the dependency arcs of the class.
//               : 
// Input         : decl - any class declaration
//               :        
// Output        : None
// Returns       : number of children
//
unsigned int MetricsCalculator::numberOfChildren(CS_ClassDeclaration& decl)
   {
   //To find the number of children for a given class, traverse its dependents
   //looking for Inheritance dependency arcs
   CS_StoragePlacement placement;
   CS_DependencyNodeIterator& iterator = decl.dependents(placement);
   CS_DependencyNode* node;
   unsigned int numberOfChildren = 0;

   for (node = iterator.firstOfKind(CS_DependencyKind::Inheritance); 
        node;
        node = iterator.nextOfKind(CS_DependencyKind::Inheritance))
      {
      numberOfChildren++;
      }
   return numberOfChildren;
   }

//////////////////////////////////////////////////////////////////////////////
//
// Function Name : numberOfVirtualFunctions
// Description   : Computes the number of user defined virtual functions declared by this class.
//               : (It does not count inherited base class virtuals)
//               : 
// Input         : classDecl - any class declaration
//               :        
// Output        : None
// Returns       : number of virtual functions
//
unsigned int MetricsCalculator::numberOfVirtualFunctions(CS_ClassDeclaration& classDecl)
   {
   CS_MemberDeclarationStore& mds = classDecl.memberDeclarationStore();
   CS_FunctionDeclaration*  fdecl;
   unsigned int numberOfVirtualFunctions = 0;
   for (CS_Declaration* decl = mds.firstDeclaration();
        decl;
        decl = mds.nextDeclaration(*decl))
       {
       if ((fdecl = decl->asFunctionDeclaration()) 
           && fdecl->mapsToASourceLocation()
           && fdecl->isVirtual())
          {
          numberOfVirtualFunctions++;  
          }
       }
   return numberOfVirtualFunctions;
   }

//////////////////////////////////////////////////////////////////////////////
//
// Function Name : numberOfPureVirtualFunctions
// Description   : Computes the number of user defined pure virtual functions declared by this class.
//               : (It does not count inherited base class virtuals)
//               : 
// Input         : classDecl - any class declaration
//               :        
// Output        : None
// Returns       : number of pure virtual functions
//
unsigned int MetricsCalculator::numberOfPureVirtualFunctions(CS_ClassDeclaration& classDecl)
   {
   CS_MemberDeclarationStore& mds = classDecl.memberDeclarationStore();
   CS_FunctionDeclaration*  fdecl;
   unsigned int numberOfPureVirtualFunctions = 0;
   for (CS_Declaration* decl = mds.firstDeclaration();
        decl;
        decl = mds.nextDeclaration(*decl))
       {
       if ((fdecl = decl->asFunctionDeclaration()) 
           && fdecl->mapsToASourceLocation()
           && fdecl->isPureVirtual())
          {
          numberOfPureVirtualFunctions++;  
          }
       }
   return numberOfPureVirtualFunctions;
   }

//////////////////////////////////////////////////////////////////////////////
//
// Function Name : numberOfNonTemplateDependentMemberFunctions 
// Description   : Iterates through a template instance and checks 
//               : each member function declaration to see if any of it's 
//                 parameters depend on a template parameter 
//               : 
// Input         : classDecl - any class declaration
//               :        
// Output        : None
// Returns       : number of template member functions that are not dependent 
//                 on the template parameters
//
unsigned int MetricsCalculator::numberOfNonTemplateDependentMemberFunctions (CS_ClassDeclaration& classDecl)
   { 
   CS_TemplateDeclarationStore* dclTds = classDecl.templateDeclarationStore();
   CS_MemberDeclarationStore& mbrStore = classDecl.memberDeclarationStore();
   CS_Declaration*                decl;
   CS_FunctionDeclaration*       fDecl;
   unsigned int total                  = 0;

   if (dclTds && dclTds->isTemplate()) // Means that this is a primary template
      {
      //Iterate through the member store
      for (decl = mbrStore.firstDeclaration();
           decl;
           decl = mbrStore.nextDeclaration(*decl))
          {
          fDecl = decl->asFunctionDeclaration();
          if (fDecl)
             {
             CS_TypeDescriptor& td = fDecl->typeDescriptor();
             CS_TypeDescriptor* returnType = td.returnType();
             if (returnType && returnType->declaration().isTemplateDependentDeclaration())
                {
                continue;
                }

             CS_FunctionParameter* parm = td.firstParameter();
             CS_bool             found  = CS_false;           
             // Check parameters
             while (parm)
                {               
                CS_TypeDescriptor& parmTd = parm->typeDescriptor();              
 
                for (CS_TypeDescriptor* nextTd = &parmTd; nextTd; nextTd = nextTd->next())
                   {
                   
                   if (nextTd->isNamedType() && nextTd->declaration().isTemplateDependentDeclaration())
                      {
                      found = CS_true;
                      break;
                      }
                   }
                
                if (found)
                   break;

                parm = parm->next();
                }
             if (!found)
                total++;
             }
          }
       }
   return total;
   }

//////////////////////////////////////////////////////////////////////////////
//
// Function Name : couplingBetweenClasses 
// Description   : Iterates through all function members of a class,
//               : count how many other classes it uses by walking  
//                 the function body 
//               :
// Input         : classDecl - any class declaration
//               :
// Output        : None
// Returns       : number of template member functions that are not dependent
//                 on the template parameters
//

unsigned int MetricsCalculator::couplingBetweenClasses(CS_ClassDeclaration& classDecl)
   {
   CS_MemberDeclarationStore& mds = classDecl.memberDeclarationStore();
   CS_Declaration* decl;
   CS_FunctionDeclaration* funcDecl;
   CS_Implementation* impl;
   CS_FunctionBody* body;
   CS_FunctionBody::BodyInfo* bodyInfo;
   CS_StorageArena arena;
   CS_StoragePlacement placement(&arena);   
   CS_LexicalBlockStatement* lbs;
   CS_PtrArray<CS_ClassDeclaration> referredClasses;
   
   class MemberFunctionCallFinder : public CS_DepthFirstVisitor
      {
   public:  
       MemberFunctionCallFinder(CS_ClassDeclaration& classDecl, 
                                CS_PtrArray<CS_ClassDeclaration>& ref) : 
                                _classDecl(classDecl), _referredClasses(ref) {}

       virtual void visitMemberExpression(CS_MemberExpression &memberExpr) 
          {
          CS_Declaration* d = memberExpr.declaration();
          CS_Declaration* f = d->asFunctionDeclaration();
          CS_ClassDeclaration* c;
          if (f)
             {
             CS_MemberDeclarationStore* store = f->enclosingDeclarationStore().asMemberDeclarationStore();
             assume (store);
             c = &store->classDeclaration();
             //Make sure we are refering to another class or not related through inheritance
             if ((c != &_classDecl) ||  !CS_DeclarationQueries::isBaseClassOf(*c, _classDecl) 
                                    ||  !CS_DeclarationQueries::isBaseClassOf(_classDecl, *c))
                {
                //Add to list of refered classes
                addClass (c);  
                }
             }

          CS_DepthFirstVisitor::visitMemberExpression(memberExpr);
          }
   private:
        void addClass(CS_ClassDeclaration* classDecl)
            {
            int i;
            CS_bool found = CS_false;
            for (i = 0; i < _referredClasses.size(); i++)
               {
               if (_referredClasses[i] == classDecl)
                  {
                  found = CS_true;
                  break;
                  }
               }
               if (!found)
                  _referredClasses.add(classDecl);
            }

        CS_ClassDeclaration& _classDecl;
        CS_PtrArray<CS_ClassDeclaration>& _referredClasses;
      };

   for (decl = mds.firstDeclaration(); decl;
        decl = mds.nextDeclaration(*decl))
      {
      // Iterate through all the function bodies
      funcDecl = decl->asFunctionDeclaration();
      if (funcDecl)
         {
         impl = funcDecl->implementation();
         if (impl)
            {
            body = impl->asFunctionBody();
            assume (body);
            body->body(bodyInfo, CS_Implementation::diagnosed, placement); 
            assume(bodyInfo);
            lbs = bodyInfo->body()->asLexicalBlockStatement();
            assume(lbs);
            // Now that we have the lexical block, walk it to see what classes are refered 
            MemberFunctionCallFinder finder(classDecl, referredClasses);
            finder.visitLexicalBlockStatement(*lbs);
            }   
         }   
      }
   return referredClasses.size();
   }
//////////////////////////////////////////////////////////////////////////////
//
// Function Name : computeMetricValue 
// Description   : Redirects a metric enum to call the corresponding 
//               : static function in MetricsCalculator class 
//               :
// Input         : decl - a declaration
//               : m    - a metric that you want to compute
//               :
// Output        : None
// Returns       : the computed value of the given metric 
//
unsigned int MetricsCalculator::computeMetricValue(CS_Declaration& decl, Metric m) 
   {
   unsigned int value = 0;
   CS_ClassDeclaration* classDecl = decl.asClassDeclaration();
   if (classDecl && decl.mapsToASourceLocation()) 
      {
      switch (m) 
        {
        case depthOfInheritanceTreeMetric:
           value = depthOfInheritanceTree(*classDecl);
           break;
        case numberOfChildrenMetric:
           value = numberOfChildren(*classDecl);
           break;
        case numberOfMemberFunctionsMetric:
           value = numberOfMemberFunctions(*classDecl);
           break;
        case numberOfVirtualFunctionsMetric:
           value = numberOfVirtualFunctions(*classDecl);
           break;
        case numberOfPureVirtualFunctionsMetric:
           value = numberOfPureVirtualFunctions(*classDecl);
           break;
        case couplingBetweenClassesMetric:
           value = couplingBetweenClasses(*classDecl);
           break;
        case numberOfNonTemplateDependentMemberFunctionsMetric:
           value = numberOfNonTemplateDependentMemberFunctions(*classDecl);
           break;
        default:
           assume(0);
           break;
        }
      }
   else
      {
      throw "Not a class declaration";
      }
   return value;
   }
  
//////////////////////////////////////////////////////////////////////////////
//
// Statistics calculation functions 
//
unsigned int StatisticsCalculator::median (Metric m)
   {
   unsigned int scoreboard [512];
   unsigned int value = 0;
   unsigned int median = 0;
   unsigned int max = 0;

   // Init the scoreboard
   for (int i = 0; i < 512; i++)
       scoreboard[i] = 0;

   // Compute stats to find median
   for (CS_Declaration* decl = _gds.firstDeclaration();
        decl;
        decl = _gds.nextDeclaration(*decl))
      {
      try 
         {
         value = MetricsCalculator::computeMetricValue(*decl, m);
         }
      catch (...)
         {
         continue;
         }   

      assume ((value >= 0) && (value <= 511));
      scoreboard[value]++;
      }

   // See which value is the median
   for (int k = 0; k < 512; k++)
      {
      if (scoreboard[k] > max)
         {
         max = scoreboard[k];
         median = k;
         }
      }

   return median;
   }

unsigned int StatisticsCalculator::minimum (Metric m)
   {
   unsigned int min = 0;
   unsigned int value = 0;

   for (CS_Declaration* decl = _gds.firstDeclaration();
        decl;
        decl = _gds.nextDeclaration(*decl))
      {
      try 
         {
         value = MetricsCalculator::computeMetricValue(*decl, m);
         }
      catch (...)
         {
         continue;
         }

      if (value < min)
         min = value;
      }

   return min;
   }

unsigned int StatisticsCalculator::maximum (Metric m) 
   {
   unsigned int max = 0;
   unsigned int value = 0;
   for (CS_Declaration* decl = _gds.firstDeclaration();
        decl;
        decl = _gds.nextDeclaration(*decl))
      {
      try 
         {
         value = MetricsCalculator::computeMetricValue(*decl, m);
         }
      catch (...)
         {
         continue;
         }
      if (value > max)
         max = value;
      }
   return max;
   }
   
//////////////////////////////////////////////////////////////////////////////
//
// Totals collection functions
//
TotalsCollector::TotalsCollector(CS_CodeStore* cs) : _gds(cs->globalDeclarationStore()), _store(cs->sourceStore())
    {
    }

int TotalsCollector::numberOfClasses()
   {
   int numberOfClasses = 0;
   for (CS_Declaration* d = _gds.firstOuterDeclaration(); d; d = _gds.nextOuterDeclaration(*d))
       {
       if (d->asClassDeclaration() && d->mapsToASourceLocation())
          {
          numberOfClasses++;
          }
       }
   return numberOfClasses;
   }

int TotalsCollector::numberOfLinesOfCode()
   {
   int numberOfLinesOfCode = 0;
   for (CS_Source* s = _store.firstSource(); s; s = s->next())
       {
       CS_PrimarySource* ps = s->asPrimarySource(); 
       if (ps)
          { 
          numberOfLinesOfCode += MetricsCalculator::numberOfLinesOfCode(*ps);
          } 
       }
   return numberOfLinesOfCode;
   }

int TotalsCollector::numberOfSpecifiedMetric(Metric m)
   {
   int sum = 0;
   for (CS_Declaration* d = _gds.firstOuterDeclaration(); d; d = _gds.nextOuterDeclaration(*d))
       {
       if (d->asClassDeclaration() && d->mapsToASourceLocation())
          {
          CS_ClassDeclaration* classDecl = d->asClassDeclaration();
          sum += MetricsCalculator::computeMetricValue(*classDecl, (Metric) m);
          }
       }
   return sum;
   }

char* getMetricsString (Metric m)
   {
   switch (m) 
      {
      case depthOfInheritanceTreeMetric:
         return "Depth of inheritance tree";
         break;
      case numberOfChildrenMetric:
         return "Number of children";
         break;
      case numberOfMemberFunctionsMetric:
         return "Number of member functions";
         break;
      case numberOfVirtualFunctionsMetric:
         return "Number of virtual functions";
         break;
      case numberOfPureVirtualFunctionsMetric:
         return "Number of pure virtual functions";
         break;
      case couplingBetweenClassesMetric:
         return "Coupling between classes";
         break;
      case numberOfNonTemplateDependentMemberFunctionsMetric:
         return "Number of non-template dependent member functions";
         break;
      default:
         assume (0);
      }
   return "";
   }
